/*   Title:  STEPISR.C                               *
 *  Author:  John Reynolds                           *
 * Purpose:  Stepper Motor Control Using             *
 *           Real Time Clock For Timing Source.      *
 *           Stepper Motor routine was adapted from  *
 *           'Computer Control Of Machines and       *
 *           Processes', by John G. Bollinger and    *
 *           Neil A. Duffie, Copyright 1988 by       *
 *           Addison-Wesley Publishing Company.      *
 *           Intended for Instructional Value only!  *
 * Copyright 1991 Magnetic Technologies Corporation  *
 *           Rochester, New York  14625              */

#include <dos.h>
#include <conio.h>
#include "stepisr.h"
#include "rtclock.h"
#include "ioutils.h"

BOOLEAN keybreak, sound_on, mech_home;
BOOLEAN eot_cw_stop, eot_ccw_stop, *eot_stop_ptr;

static float P, V, A;
static int N;
static unsigned int StepCount;

/*-------------------------------*/
void Rtc_profile_segment_move(void)
{
  if (RtcCounter < N) /* don't take too many steps */
    {
      RtcCounter++;
      P += V;
      V += A;
      if ((int)P) /* +1 or -1 overflow */
      {
        if (sound_on)
        {
          sound(500);
          nosound();
        } /* if sound_on */
          else
        {
          turn_on_bit(STEP); /* take step */
          turn_off_bit(STEP); /* reset step */
        } /* sound off */

         StepCount++;
        P -= (int)P;
        read_io(StepCount); /* take all readings */
      } /* if overflow */
  } /* if (RtcCount < N) */

  if (kbhit()) keybreak = TRUE;
 } /*  Rtc_profile_segment_move() */

/*--------------------------------------*/
void profile_segment_move(int clk_freq_hz)
{
  if ( *eot_stop_ptr || keybreak) return;

  RtcCounter = StepCount = 0;
  NewRtcVec = Rtc_profile_segment_move;

  switch(clk_freq_hz)
  {
    case  128: EnableRtcInts(Hz128);
               break;
    case 1024: EnableRtcInts(Hz1024);
               break;
    case 4096: EnableRtcInts(Hz4096);
               break;
   } /* switch */

  /* wait for interrupt routine to finish move */
  while (!( (RtcCounter >= N) || keybreak ||
             *eot_stop_ptr));

  DisableRtcInts();
  while (kbhit()) getch(); /* clear buffer */
} /* profile_segment_move() */

/*-----------------------*/
unsigned int move_stepper (
   long int steps,
   float velocity,
   float accel,
   int clk_freq_hz        )
{
  int i, sum, N2;
  float P2, A1, N1;
  unsigned int steps_moved = 0;

  keybreak = FALSE;
  /* validate frequency request */
   if (   clk_freq_hz != 128
       && clk_freq_hz != 1024
       && clk_freq_hz != 4096 ) return 0;

  /* make sure motor safe to move */
   eot_stop_ptr =
     (velocity < 0) ? &eot_ccw_stop : &eot_cw_stop;
   read_io(0); /* read eot_stop status */
   if (*eot_stop_ptr) return 0;

  /* Stop any potential sign problems -
     only velocity can be negative */
   steps = (steps<0) ? -steps : steps;
   accel = (accel<0) ? -accel : accel;

  /* Calculate Velocity Profile */
   V = velocity / clk_freq_hz;  /* steps/tick */
   A1 = accel / clk_freq_hz / clk_freq_hz; /* steps/tick/tick */
   N1 = (int)( V / A1 );
   N1 = (N1<0) ? -N1 : N1;
   sum = 0;
   for (i=1; i<=N1; i++) sum += i;
   N2 = (int)( (steps - 2*(int)(A1*sum)) / (N1*A1));

   /* Get ready to go */
     turn_on_bit(DIRECTION); /* clockwise motion */
     if (V<0) /* counterclockwise motion */
     {
       A1 = -A1;
       turn_off_bit(DIRECTION);
     } /* if counterclockwise */
     P = V = 0.0;
   /* Accelerate */
     A = A1;
     N = N1;
     profile_segment_move(clk_freq_hz);
     steps_moved += StepCount;
   /* Constant velocity */
     A = 0.0;
     N = N2;
     profile_segment_move(clk_freq_hz);
     steps_moved += StepCount;
   /* Decelerate */
     A = -A1;
     N = N1;
     profile_segment_move(clk_freq_hz);
     steps_moved += StepCount;
   return(steps_moved);
} /* move_stepper() */
