/* BTN_FNCS: 3-D Buttons Support Functions

   Copyright (c) 1992 Springtime Software
   All Rights Reserved

   Uses owner-draw 3-D buttons with special text in their captions
   describing 3-D text and graphics effects.

   User Pushbutton caption fields:

	@f ............................ 3-D Frame (must come first)
	@b<name> ...................... Draw Named Bitmap With Key Bitmap
									(see below)
	@t<text string> ............... Draw Text String
	@<modifiers>t<text string> .... Draw Text String

   The modifiers are:
		2 ............. 2-D (Floating) Text or Graphic
		e ............. Engraved
		r ............. Raised (Embossed)
		u ............. Up (At the Top)
		d ............. Down (At the Bottom)

   Each bitmap has a corresponding "key" bitmap called <name>_3d,
   where <name> is the name of the bitmap.  In this key bitmap colors
   indicate the 3-D effects, as follows:

   Key Colors:

		Light Gray ........ Background that moves with button
		Red ............... Flat areas that move
		Green ............. Engraved areas that move
		Blue .............. Raised (Embossed) areas that move

		Gray .............. Background that "floats" above button
		Dark Red .......... Flat areas that float
		Dark Green ........ Engraved areas that float
		Dark Blue ......... Raised areas that float

        White ............. Transparent (clear) areas
*/

#include <windows.h>
#include "btn_fncs.h"

/* Standard declarations */
#define EQ ==
#define NE !=
#define NUL '\0'
#define STRCMP lstrcmp

/* Global declarations */
static LPSTR			app_name;
static HANDLE			instance;
static ERRPROC far *	fatal;

/* For debugging: Display given DC+bitmap in center of screen */
void	show(HDC dc_in)
	{
	HDC			dc;

    dc = GetDC(NULL);
	BitBlt(dc, 300, 200, 100, 100, dc_in, 0, 0, SRCCOPY);
	ReleaseDC(NULL, dc);
	} /* show */

/* 3-D Pushbutton Parameters */
#define			DELIM '@' /* Field delimiter in 3-D control string */
#define			MOVEMENT 2			/* Of button when it is pushed, */
									/* in units of border thickness */
#define			SHADOW_OFFSET 1		/* Additional offset of shadows of */
									/* 2d areas onto 3d background */

/* Maximum length of button caption text */
#define	MAXTEXTLEN 512

/* Standard 16 VGA Colors */
#define			COLOR_BLACK		RGB(0, 0, 0)
#define			COLOR_DKRED		RGB(128, 0, 0)
#define			COLOR_DKGREEN	RGB(0, 128, 0)
#define			COLOR_DKYELLOW	RGB(128, 128, 0)
#define			COLOR_DKBLUE	RGB(0, 0, 128)
#define			COLOR_DKMAGENTA	RGB(128, 0, 128)
#define			COLOR_DKCYAN	RGB(0, 128, 128)
#define			COLOR_LTGRAY	RGB(192, 192, 192)
#define			COLOR_GRAY		RGB(128, 128, 128)
#define			COLOR_RED		RGB(255, 0, 0)
#define			COLOR_GREEN		RGB(0, 255, 0)
#define			COLOR_YELLOW	RGB(255, 255, 0)
#define			COLOR_BLUE		RGB(0, 0, 255)
#define			COLOR_MAGENTA	RGB(255, 0, 255)
#define			COLOR_CYAN		RGB(0, 255, 255)
#define			COLOR_WHITE		RGB(255, 255, 255)

/* This nonstandard color looks too close to black to be usable: */
/*#define 		COLOR_DKGRAY 	RGB(64, 64, 64) */

/* Shadows of "floating" areas on moving areas are always gray */
#define			COLOR_SHADOW	COLOR_GRAY

/* 3-D effects colors used in the <name>_3d "key" bitmap */
#define			EFF_CLEAR		COLOR_WHITE
#define			EFF_3D_BKGND	COLOR_LTGRAY
#define			EFF_3D_FLAT		COLOR_RED
#define			EFF_3D_ENGRAVED	COLOR_GREEN
#define			EFF_3D_RAISED	COLOR_BLUE
#define			EFF_2D_BKGND	COLOR_GRAY
#define			EFF_2D_FLAT		COLOR_DKRED
#define			EFF_2D_ENGRAVED	COLOR_DKGREEN
#define			EFF_2D_RAISED	COLOR_DKBLUE

/* Effects modifiers bits (reflects chars after delimiter in caption text */
/* control string) */
typedef struct
	{
	UINT	mod_2 :1;
	UINT	mod_e :1;
	UINT	mod_r :1;
	UINT	mod_u :1;
	UINT	mod_d :1;
	UINT	pad  :11;
    }			
				MOD_TYPE;
MOD_TYPE		modifiers;

/* Raster Operation (ROP) codes for patblt() and bitblt() */
#define			PATANDDEST 0x00A000C9L /* ROP code for Brush & Destination */
#define			PATORDEST 0x00FA0089L /* ROP code for Brush | Destination */
#define			MASKEDBRUSH 0x00E20746L /* ROP code for Source & Brush | */
										/*     ~ Source & Destination */
#define			PATSRCEQ 0x00C3006AL /* ROP code for Brush Equals Source */
#define			DSTERASE 0x00220326L /* ROP code for ~Source & Destination */

/* 3-D global declarations */
#ifndef max
#define max(a, b) ((a) < (b) ? (b) : (a))
#define min(a, b) ((a) < (b) ? (a) : (b))
#endif
#define			WIDTH(lprect) ((lprect)->right - (lprect)->left)
#define			HEIGHT(lprect) ((lprect)->bottom - (lprect)->top)
#define			IDLEN 128
typedef enum {SAME_COLOR, DARKER, BRIGHTER, SHADOW} BRIGHTNESS_TYPE;
typedef enum {NO_MOVE=-MOVEMENT, MOVE=0, END=100} MOVE_TYPE;
HDC				dc_up, dc_down, dc_button, dc_bitmap, dc_key;
HBITMAP			bitmap_up, bitmap_down;
RECT			up, down; 	/* Display surface areas within button */
LPSTR			field; /* Pointer to current caption field */
HBITMAP			bitmap, bitmap_key;
HWND			button;
char			caption[MAXTEXTLEN];
UINT			text_format;
int				width, height; /* Of entire button window */
/*int				movement; * Motion of image, up/down */
int				x_border, y_border, x, y, layer, dialog_level;
COLORREF		color_btntext;

/* Delete a GDI object, with error check */
static void	del(HANDLE object)
	{
	if (!DeleteObject(object))
		fatal("attempted to delete a GDI object 0x%lX that was "
			"already deleted or is still selected in a DC",
			(long)(void FAR *)object);
	} /* del */

/* Create a Memory DC containing a new or existing bitmap,
   monochrome or color. If a bitmap is to be created, uses global
   width and height values.  Assumes that dc_button is an active
   color window DC and that dc_up is an active memory DC. */
static void	create_dc_with_bitmap(HDC * dc_out, HBITMAP * bitmap,
							  BOOL in_color)
	{
	if (!dc_out || !bitmap)
		fatal("a ref arg is NULL", NULL);

	/* Create memory DC */
	*dc_out = CreateCompatibleDC(in_color ? dc_button : dc_up);
	if (!*dc_out)
		fatal("cannot create a device context", NULL);

	/* Create bitmap, if needed, and select into memory DC */
	if (!*bitmap)
		*bitmap = CreateCompatibleBitmap(in_color ? dc_button : *dc_out,
			width, height);
	if (!*bitmap)
		fatal("cannot create a bitmap", NULL);
	SelectObject(*dc_out, *bitmap);

	/* Initialize memory DC */
	SetBkMode(*dc_out, TRANSPARENT); /* For painting text over graphics */
	return;
	} /* create_dc_with_bitmap */

/* Draw a rectangle of the given color in the given DC */
static void	color_area(HDC dc, LPRECT lpr, COLORREF color)
	{
	static LOGBRUSH		br;
	HBRUSH				brush;

	/* Create solid brush and select it into the DC */
	br.lbStyle = BS_SOLID;
	br.lbColor = color;
	brush = CreateBrushIndirect(&br);
	SelectObject(dc, brush);

	/* Paint the brush (pattern) color */
	/* (Slightly faster than the FillRect function) */
	PatBlt(dc, lpr->left, lpr->top,	WIDTH(lpr), HEIGHT(lpr), PATCOPY);

	/* Deselect and delete the brush */
	del(SelectObject(dc, GetStockObject(NULL_BRUSH)));
	} /* color_area */

/* Draw a one-pixel rectangle with given 3-D shading as part of a button */
/* frame or sides, at the given rectangle location */
static void	draw_frame_line(HDC dc, COLORREF lt, COLORREF rb, LPRECT lpr)
	{
	HPEN		pen_lt, pen_rb;

	/* Init */
	pen_lt = CreatePen(PS_SOLID, 1, lt);
	pen_rb = CreatePen(PS_SOLID, 1, rb);

	/* Draw the four sublines that make one frame line */
	MoveTo(dc, lpr->left, lpr->bottom - 2); /* Start at LB corner */
	SelectObject(dc, pen_lt);
	LineTo(dc, lpr->left, lpr->top); /* L edge */
	LineTo(dc, lpr->right - 1, lpr->top); /* T edge */
	SelectObject(dc, pen_rb);
	LineTo(dc, lpr->right - 1, lpr->bottom - 1); /* R edge */
	LineTo(dc, lpr->left - 1, lpr->bottom - 1); /* B edge */

	/* Done */
	SelectObject(dc, GetStockObject(BLACK_PEN));
	del(pen_lt);  del(pen_rb);
	} /* draw_frame_line */

/* Draw a 3-D button frame at the given rectangle location on up */
/* and down bitmaps (the rect is shrunk as a side effect) */
static void	draw_3d_frame(LPRECT lpr)
	{
	draw_frame_line(dc_up, COLOR_WHITE, COLOR_GRAY, lpr);
	draw_frame_line(dc_down, COLOR_WHITE, COLOR_GRAY, lpr);
	InflateRect(lpr, -1, -1);
	draw_frame_line(dc_up, COLOR_LTGRAY, COLOR_LTGRAY, lpr);
	draw_frame_line(dc_down, COLOR_LTGRAY, COLOR_LTGRAY, lpr);
	InflateRect(lpr, -1, -1);
	draw_frame_line(dc_up, COLOR_GRAY, COLOR_WHITE, lpr);
	draw_frame_line(dc_down, COLOR_GRAY, COLOR_WHITE, lpr);
	InflateRect(lpr, -1, -1);
	} /* draw_3d_frame */

/* Draw a two-pixel frame (the inner line is just skipped) around a */
/* button at the given rectangle location on up and down bitmaps */
/* (the rect is shrunk as a side effect). */
/* The actual thickness of this "default button" frame is decided later, */
/* at actual button painting time. */
static void	draw_frame(LPRECT lpr)
	{
	draw_frame_line(dc_up, COLOR_BLACK, COLOR_BLACK, lpr);
	draw_frame_line(dc_down, COLOR_BLACK, COLOR_BLACK, lpr);
	InflateRect(lpr, -1, -1);
	draw_frame_line(dc_up, COLOR_BLACK, COLOR_BLACK, lpr);
	draw_frame_line(dc_down, COLOR_BLACK, COLOR_BLACK, lpr);
	InflateRect(lpr, -1, -1);
	} /* draw_frame */

/* Draw the sides of a 3-D button at the given rectangle location on up */
/* and down bitmaps (the rectangles are shrunk, and the button surface */
/* is painted light gray as side effects) */
static void	draw_sides(LPRECT lp_up, LPRECT lp_down)
	{
	/* Up state */
	draw_frame_line(dc_up, COLOR_WHITE, COLOR_GRAY, lp_up);
	InflateRect(lp_up, -1, -1);
	draw_frame_line(dc_up, COLOR_WHITE, COLOR_GRAY, lp_up);
	InflateRect(lp_up, -1, -1);

	/* Default lt gray color */
	color_area(dc_up, lp_up, COLOR_LTGRAY);
	color_area(dc_down, lp_down, COLOR_LTGRAY);

	/* Down state */
	SetRect(lp_down, lp_down->left, lp_down->top,
		lp_down->right + 1, lp_down->bottom + 1); /* Locate shadow */
	draw_frame_line(dc_down, COLOR_SHADOW, COLOR_BLACK, lp_down);
	SetRect(lp_down,
		lp_down->left + MOVEMENT + 2, lp_down->top + MOVEMENT + 2,
		lp_down->right - 1, lp_down->bottom - 1); /* Locate surface */
	} /* draw_sides */

/* Gray the picture inside the rectangle in the given 3-D button, */
/* to indicate that this control is disabled */
static void    gray_area(HDC dc, LPRECT lpr)
	{
	HBRUSH			brush;
	HBITMAP			gray_bitmap;

	/* AND with solid Light Gray brush: Makes colors darker */
	SelectObject(dc, GetStockObject(LTGRAY_BRUSH));
	PatBlt(dc, lpr->left, lpr->top,	WIDTH(lpr), HEIGHT(lpr),
		PATANDDEST /* P & D */);

	/* OR with Light Gray/Black checkerboard to make the picture appear */
	/* lighter again with gray dots */
	gray_bitmap = LoadBitmap(instance, "gray_brush");
	if (!gray_bitmap)
		fatal("gray brush bitmap could not be loaded", NULL);
	brush = CreatePatternBrush(gray_bitmap);
	SelectObject(dc, brush);
	PatBlt(dc, lpr->left, lpr->top,	WIDTH(lpr), HEIGHT(lpr),
		PATORDEST /* P | D */);
	SelectObject(dc, GetStockObject(WHITE_BRUSH));
	del(brush);
	del(gray_bitmap);
	} /* gray_area */

/* Paint a given bitmap into the given button window dc, */
/* which is assumed to have a compatible bitmap already selected. */
/* (See Programming Windows, Petzold, 2nd Ed, p. 621, for a more */
/* general version) */
static void	paint_bitmap(HDC dc, HBITMAP bitmap)
	{
	HDC				dc_mem;
	BITMAP			bm;

	dc_mem = CreateCompatibleDC(dc);
	if (!dc_mem)
    	fatal("cannot create a device context", NULL);
	SelectObject(dc_mem, bitmap);
	GetObject(bitmap, sizeof(BITMAP), (LPSTR)&bm);
	BitBlt(dc, 0, 0, bm.bmWidth, bm.bmHeight,
		dc_mem, 0, 0, SRCCOPY);
	DeleteDC(dc_mem);
	} /* paint_bitmap */

/* Return a temporary rectangle representing the location of the */
/* extra outer "default" frame line on a given 3-D button */
static LPRECT	default_frame_of(HWND button)
	{
	static RECT		surface;
	int				frame;

	GetClientRect(button, &surface);
	frame = (short)GetProp(button, "3Df");
	InflateRect(&surface, - frame, - frame);
    return &surface;
	} /* default_frame_of */

/* Return a temporary rectangle representing the surface (display) */
/* area of a given 3-D button */
static LPRECT	surface_of(HWND button, BOOL down)
	{
	LPRECT			lpsurface;

	lpsurface = default_frame_of(button);
	InflateRect(lpsurface, - 4, - 4);
	if (down)
    	OffsetRect(lpsurface, 2, 2);
	return lpsurface;
	} /* surface_of */

/* Locate the next field and mark the end of the current field string: */
/* change next 3-D delimiter to a NUL, so a field in a caption string */
/* can be used as a NUL-delimited string */
BOOL	marked;
LPSTR	c1; /* Pointer to start of next field (or, if none, to NUL) */
static void	mark_eos(void)
	{
	marked = FALSE;
	for (c1 = field; *c1; c1++)
		{
		if (*c1 EQ DELIM)
			{
			marked = TRUE;
			*c1 = NUL;
			break;
			} /* Found delimiter */
		} /* Scan string */
	} /* mark_eos */

/* Undo the effect of the last call to mark_eos (these must be called */
/* in non-nested pairs) and skip to the next field */
static void	unmark_eos(void)
	{
	if (marked)
		*c1 = DELIM;
	field = c1;
	} /* unmark_eos */

/* Text Display Part */

/* Draw a text image on a given DC+bitmap (text_format is global) */
static void	draw_text_effect(HDC dc, LPRECT lpr,
						 	 COLORREF color, int offset)
	{
	OffsetRect(lpr, offset * x_border, offset * y_border);
	SetTextColor(dc, color);
	DrawText(dc, field, -1, lpr, text_format);
	OffsetRect(lpr, - offset * x_border, - offset * y_border);
	} /* draw_text_effect */

/* Draw a single text image on up and down bitmaps, given an offset (in */
/* pixels) of the image from its normal position (used to create highlights
 * and shadows) and the amount of movement (in border thicknesses) between
 * the up and down states */
static void	draw_text_image(COLORREF color, int offset)
	{
	draw_text_effect(dc_up, &up, color, offset);
	if (modifiers.mod_2) /* This text is 2-D (it doesn't move) */
    	{
		draw_text_effect(dc_down, &down, COLOR_SHADOW,
			offset + SHADOW_OFFSET); /* Shadow on 3-D surface */
		draw_text_effect(dc_down, &down, color,
			offset - MOVEMENT); /* Remove normal 3-D movement */
		}
    else
		draw_text_effect(dc_down, &down, color, offset);
	} /* draw_text_image */

/* Draw 3-D text */
static void	draw_text(void)
	{
	/* Set up for text drawing */
	text_format = DT_NOCLIP | DT_SINGLELINE | DT_TABSTOP | DT_CENTER;

	if (modifiers.mod_u) /* Place this text at top */
		text_format |= DT_TOP;
	else if (modifiers.mod_d) /* Place this text at bottom */
		text_format |= DT_BOTTOM;
	else
		text_format |= DT_VCENTER; /* Default is vertically centered */
	if (modifiers.mod_r) /* Raised */
    	{
		draw_text_image(COLOR_WHITE, -1); 	/* Highlight */
		draw_text_image(COLOR_GRAY, 1);		/* Shadow */
		}
	if (modifiers.mod_e) /* Engraved Text */
		draw_text_image(COLOR_WHITE, 1);	/* Highlight */
	draw_text_image(color_btntext, 0);		/* Text Itself */
	} /* draw_text */

/* Graphics Display Part */

/* Move and draw a given effect DC+bitmap, transparently and with a */
/* given offset, onto the given up or down DC+bitmap */
static void	draw_mask(HDC dc_dest, LPRECT lpr_dest, HDC dc_effect, HDC dc_bw,
					  int offset)
	{
	/* Turn to black the effect areas on the dest DC+bitmap */
	BitBlt(dc_dest,
		lpr_dest->left + offset * x_border,
		lpr_dest->top + offset * y_border,
		width, height,
		dc_bw, 0, 0, DSTERASE /* ~S & D */);

	/* Copy (OR) the effect DC+bitmap to the dest DC+bitmap */
	BitBlt(dc_dest,
		lpr_dest->left + offset * x_border,
		lpr_dest->top + offset * y_border,
		width, height,
		dc_effect, 0, 0, SRCPAINT /* S | D */);
	} /* draw_mask */

/* Draw one 3-D effect: Mask the bitmap and copy the result
 * to the up/down DCs/bitmaps, changing colors and positions
 * to simulate illumination, shading, shadows, and movement.
 * Uses global width and height. */
static void	draw_effect(COLORREF effect_color, BRIGHTNESS_TYPE	brightness,
						int	offset, MOVE_TYPE movement)
	{
	HDC				dc_effect;
	HBITMAP			bitmap_effect;
	HBRUSH			brush;
	HDC				dc_bw;
	HBITMAP			bitmap_bw;

	/* Note: This assumes that the effect color is a COLORREF, not a
	 * bitmap color; otherwise...
	 * we would need to do a color translation using the color table
	 * stored inside the given bitmap, which requires getting the GDI
	 * segment number, somehow constructing a pointer to the bitmap
	 * (which is in GDI's default local heap), then locating the
	 * RGBQUAD color table.  Windows has no documented way to do this. */

	/* Create a color DC+bitmap for generating the 3-D effect areas */
	bitmap_effect = NULL;
	create_dc_with_bitmap(&dc_effect, &bitmap_effect, TRUE);

	/* Add a solid brush of the given effect color */
	brush = CreateSolidBrush(effect_color);
	SelectObject(dc_effect, brush);

	/* Make only the effects areas white */
	BitBlt(dc_effect, 0, 0, width, height, dc_key, 0, 0, PATSRCEQ);

	/* Convert effects to monochrome (making all colors other than white */
	/* turn into black) */
	bitmap_bw = NULL;
	create_dc_with_bitmap(&dc_bw, &bitmap_bw, FALSE);
	BitBlt(dc_bw, 0, 0, width, height, dc_effect, 0, 0, SRCCOPY);

	/* Copy the user DC+bitmap to the effect DC+bitmap */
	BitBlt(dc_effect, 0, 0, width, height, dc_bitmap, 0, 0, SRCCOPY);

	/* Change the brightness of the effects areas */
	switch (brightness)
		{
        case SHADOW:
			SelectObject(dc_effect, GetStockObject(GRAY_BRUSH));
			BitBlt(dc_effect, 0, 0, width, height,
				dc_bw, 0, 0, MASKEDBRUSH /* S & P | ~S & D */);
			break;
		case DARKER:
			SelectObject(dc_effect, GetStockObject(LTGRAY_BRUSH));
			PatBlt(dc_effect, 0, 0, width, height, PATANDDEST /* P & D */);
			break;
		case SAME_COLOR:
        	break;
		case BRIGHTER:
			SelectObject(dc_effect, GetStockObject(LTGRAY_BRUSH));
			PatBlt(dc_effect, 0, 0, width, height, PATORDEST /* P | D */);
			break;
		default:
			fatal("no such case", NULL);
		} /* Switch on brightness */

	/* Turn to black the non-effect areas on the effect DC+bitmap */
	BitBlt(dc_effect, 0, 0,	width, height,
		dc_bw, 0, 0, SRCAND /* S & D */);

	/* Copy the effect DC+bitmap transparently onto the up and down */
	/* DCs+bitmaps */
    if (brightness NE SHADOW)
		draw_mask(dc_up, &up, dc_effect, dc_bw, offset);
	draw_mask(dc_down, &down, dc_effect, dc_bw, offset + (int)movement);

	/* Cleanup */
	DeleteDC(dc_bw);
	del(bitmap_bw);
	DeleteDC(dc_effect);
	del(bitmap_effect);
	del(brush);
	} /* draw_effect */

/* Draw given bitmap on up and down DCs+bitmaps according to given */
/* "key" bitmap */
static void	draw_bitmap(void)
	{
	char				key_name[IDLEN]; /* Name of "key" bitmap */

	/* Table of painted 3-D graphics effects */
	typedef struct
		{
		COLORREF		effect_color;
		BRIGHTNESS_TYPE	brightness;
		int				offset;
		MOVE_TYPE		movement;
		}				EFFECT_TYPE;
	EFFECT_TYPE *		e;
	static EFFECT_TYPE	effect[] =
		{
			/* 3-D (Moving) effects */

			/* 3-D background area */
			{EFF_3D_BKGND, SAME_COLOR, 0, MOVE},

			/* Shadows from 2-D areas onto 3-D background */
			{EFF_2D_FLAT, SHADOW, SHADOW_OFFSET, MOVE},
			{EFF_2D_RAISED, SHADOW, SHADOW_OFFSET, MOVE},
			{EFF_2D_ENGRAVED, SHADOW, SHADOW_OFFSET, MOVE},

			/* Flat areas */
			{EFF_3D_FLAT, SAME_COLOR, 0, MOVE},

			/* Engraved areas */
			{EFF_3D_ENGRAVED, BRIGHTER, 1, MOVE},
			{EFF_3D_ENGRAVED, SAME_COLOR, 0, MOVE},

			/* Raised (embossed) areas */
			{EFF_3D_RAISED, BRIGHTER, -1, MOVE},
			{EFF_3D_RAISED, DARKER, 1, MOVE},
			{EFF_3D_RAISED, SAME_COLOR, 0, MOVE},

			/* 2-D (Floating, non-moving) effects */

			{EFF_2D_BKGND, SAME_COLOR, 0, NO_MOVE},
			{EFF_2D_FLAT, SAME_COLOR, 0, NO_MOVE},
			{EFF_2D_ENGRAVED, BRIGHTER, 1, NO_MOVE},
			{EFF_2D_ENGRAVED, SAME_COLOR, 0, NO_MOVE},
			{EFF_2D_RAISED, BRIGHTER, -1, NO_MOVE},
			{EFF_2D_RAISED, DARKER, 1, NO_MOVE},
			{EFF_2D_RAISED, SAME_COLOR, 0, NO_MOVE},

			/* End */
			{0, SAME_COLOR, 0, END}
    	};

	/* Read bitmap resource from the .EXE file into its own DC */
	bitmap = LoadBitmap(instance, field);
	if (!bitmap)
		fatal("bitmap %s could not be found", (long)field);
	create_dc_with_bitmap(&dc_bitmap, &bitmap, TRUE);

	/* Same for the "key" bitmap <name>_3d */
	wsprintf(key_name, "%s_3d", field);
	bitmap_key = LoadBitmap(instance, key_name);
	if (!bitmap_key)
		fatal("\"key\" bitmap %s could not be found",
                                (long)(void FAR *)key_name);
	create_dc_with_bitmap(&dc_key, &bitmap_key, TRUE);

	/* Simulate each 3-D effect */
	for (e = effect; e->movement NE END; e++)
		draw_effect(e->effect_color, e->brightness, e->offset, e->movement);

	/* Delete the loaded bitmap and its DC, same for key */
	DeleteDC(dc_bitmap);
	dc_bitmap = NULL;
	del(bitmap);
	bitmap = NULL;
	DeleteDC(dc_key);
	dc_key = NULL;
	del(bitmap_key);
	bitmap_key = NULL;
	} /* draw_bitmap */

/* Draw current field on up and down views */
static void	draw_field(void)
	{
	if (*field NE DELIM)
		fatal("3-D button format invalid: %s",
                            (long)(void FAR *)caption);
	field++; /* Skip delimiter */

	/* Look for modifier letters */
	*(short *)&modifiers = 0;
	for (; *field; field++)
   		{
		switch (*field)
			{   									/* Modifiers: */
			case '2': modifiers.mod_2 = 1; continue;/* 2-D (Floating) */
			case 'e': modifiers.mod_e = 1; continue;/* Engraved */
			case 'r': modifiers.mod_r = 1; continue;/* Raised (Embossed) */
			case 'u': modifiers.mod_u = 1; continue;/* Up (text at top) */
			case 'd': modifiers.mod_d = 1; continue;/* Down (text at bottom) */
			} /* Switch on letter */
		break;
		} /* Loop on modifier letters */

    if (!*field)
		fatal("3-D button format invalid: %s",
                            (long)(void FAR *)caption);

	/* Switch on field letter and locate to and delimit its string */
	/* subfield, if any */
	mark_eos(); /* Delimit string */
	switch (*field++)
		{
		case 'f':	/* Frame (handled already) */
			break;
		case 't':	/* Text */
			draw_text();
			break;
		case 'b':	/* Bitmap, with effects colors specified in */
					/* "key" bitmap <name>_3d */
			if (*(short *)&modifiers NE 0)
				fatal("3-D button format invalid: %s",
                                    (long)(void FAR *)caption);
			draw_bitmap();
            break;
		default:
			fatal("3-D button format invalid: %s",
                            (long)(void FAR *)caption);
		} /* Switch on field letter */
    unmark_eos();
	} /* draw_field */

/* Draw the complete up and down bitmaps for a given 3-D button
 * at dialog initialization time (text caption is global and has
 * already been read) */
static void	draw_3d_button(HWND button_arg)
	{
    UINT		style;

	/* Make button window global */
    button = button_arg;

	/* If this button has already been analyzed, error */
	if (GetProp(button, "3Du"))
		fatal("a button already has 3-D properties %s at a call "
			"to draw_3d_button", (long)(void FAR *)caption);

	/* Disable double-clicking (makes no sense for buttons anyway) */
	style = GetClassWord(button, GCW_STYLE);
	SetClassWord(button, GCW_STYLE, style & ~CS_DBLCLKS);

	/* Start drawing at the button border */
	GetClientRect(button, &up);
	width = WIDTH(&up); /* Global */
	height = HEIGHT(&up); /* Global */

	/* Create a DC that will be used for color information */
	dc_button = GetDC(button); /* Global */

	/* Create Up and Down memory device contexts and bitmaps and
	 * store the bitmaps as window properties of the button */
	bitmap_up = bitmap_down = NULL;
	create_dc_with_bitmap(&dc_up, &bitmap_up, TRUE);
	create_dc_with_bitmap(&dc_down, &bitmap_down, TRUE);
	SetProp(button, "3Du", bitmap_up);
	SetProp(button, "3Dd", bitmap_down);

	/* If requested, draw a 3-D frame (<delim>f must come first) */
	if (caption[1] EQ 'f')
		draw_3d_frame(&up);

	/* Store offset to outermost "default" frame as a button property */
	SetProp(button, "3Df", (HANDLE)(up.top)); /* Frame thickness */

	/* Draw a black button frame 2 pixels thick (will be thinned to
	 * one pixel when the button does not have the input focus, when
	 * button is painted) */
	draw_frame(&up);

	/* Split the drawing rectangle into two (for up and down states) */
    CopyRect(&down, &up);

	/* Draw the button sides */
	draw_sides(&up, &down);

	/* Here begins the button surface; record its height and width as
	 * global variables */
	width = WIDTH(&up); /* Global */
	height = HEIGHT(&up); /* Global */

	/* Draw each field in the button caption (title) text */
	for (field = caption; *field;)
    	draw_field();

	/* Free all GDI objects except the up and down bitmaps */
	DeleteDC(dc_up);
	dc_up = NULL;
	DeleteDC(dc_down);
	dc_down = NULL;
	ReleaseDC(button, dc_button);
	button = NULL;
	dc_button = NULL;
	} /* draw_3d_button */

/* Overall 3-D initializing and termination */

/* Initialize or terminate a single 3-D button */
FARPROC			for_each_childi;
static BOOL CALLBACK for_each_child(HWND child, LPARAM lParam)
	{
	if (lParam EQ 0)
		{
		/* Initialize a 3-D button */

		/* If this child window is not a pushbutton, ignore it */
		GetClassName(child, caption, MAXTEXTLEN);
		if (STRCMP(caption, "Button"))
			return TRUE; /* Continue scanning */

		/* If the caption text starts with the 3-D field delimiter,
		 * indicating that this is a 3-D button, analyze the fields,
		 * and construct the up and down bitmaps as window properties */
		GetWindowText(child, caption, MAXTEXTLEN);
		if (caption[0] EQ DELIM)
			draw_3d_button(child);
		} /* Initializing */
	else
		{
		/* Terminate a 3-D button */
		HBITMAP				bitmap;

		/* Free all memory, if any, used by this control */
		bitmap = RemoveProp(child, "3Du");
		if (bitmap)
			del(bitmap);
		bitmap = RemoveProp(child, "3Dd");
		if (bitmap)
			del(bitmap);
		RemoveProp(child, "3Ds");
		RemoveProp(child, "3Df");
		} /* Terminating */
	return TRUE; /* Continue scanning */
	} /* for_each_child */

/* Initialize or terminate all 3-D buttons in a dialog box */
/* (call from WM_INITDIALOG and WM_DESTROY) */
void	scan_3d_buttons(HWND dlg, LPARAM pass)
	{
	/* Scan all child windows */
	for_each_childi = MakeProcInstance((FARPROC)for_each_child, instance);
	EnumChildWindows(dlg, (WNDENUMPROC)for_each_childi, pass);
	FreeProcInstance(for_each_childi);
	} /* scan_3d_buttons */

/* Initialize 3-D button facility (call from start of application) */
void	init_3d(LPSTR app_name1, HANDLE instance1, ERRPROC fatal1)
	{
	app_name = app_name1;
	instance = instance1;
    fatal = fatal1;
	x_border = GetSystemMetrics(SM_CXBORDER);
	y_border = GetSystemMetrics(SM_CYBORDER);
	color_btntext = GetSysColor(COLOR_BTNTEXT);
	} /* init_3d */

/* Terminate 3-D button facility (call from end of application) */
void	end_3d(void)
	{
	/* Delete any temporary objects in existence */
	if (dc_up)
		DeleteDC(dc_up);
	if (dc_down)
		DeleteDC(dc_down);
    if (dc_button)
		ReleaseDC(button, dc_button);
	if (dc_bitmap)
		DeleteDC(dc_bitmap);
	if (bitmap)
		del(bitmap);
	if (dc_key)
		DeleteDC(dc_key);
	if (bitmap_key)
		del(bitmap_key);
	} /* end_3d */

/* Paint function */

/* Paint a 3-D button in a given new state (called by a dialog box */
/* WM_DRAWITEM message on behalf of a single button in the box) */
void	paint_3d_button(DWORD lParam)
	{
	LPDRAWITEMSTRUCT	lpdi;
	HWND				button;
	UINT				new_state;
    HBITMAP				bitmap_up, bitmap_down;
	LPRECT				lp_surface;
	COLORREF			color_window;

	lpdi = (LPDRAWITEMSTRUCT)lParam;
	button = lpdi->hwndItem;
	bitmap_up = (HBITMAP)GetProp(button, "3Du");
	bitmap_down = (HBITMAP)GetProp(button, "3Dd");
	new_state = lpdi->itemState;
	color_window = GetSysColor(COLOR_WINDOW);

	/* If this owner-draw control is NOT a 3-D button, error */
	if (!bitmap_up)
		fatal("a control marked in the .RC file as Owner-Draw "
			"has no 3-D caption", NULL);

	/* Paint up or down bitmap (according to selection state) onto */
	/* the 3-D button */
	if (new_state & ODS_SELECTED)
		paint_bitmap(lpdi->hDC, bitmap_down);
	else
		paint_bitmap(lpdi->hDC, bitmap_up);

	/* If the button is disabled, gray its surface */
	if (new_state & ODS_DISABLED)
		gray_area(lpdi->hDC,
			surface_of(button, new_state & ODS_SELECTED));

	/* Thicken the frame when the button has input focus */
	lp_surface = default_frame_of(button);
	if (new_state & ODS_FOCUS)
		draw_frame_line(lpdi->hDC, COLOR_BLACK, COLOR_BLACK, lp_surface);
	else
		draw_frame_line(lpdi->hDC, color_window, color_window, lp_surface);

	/* Focus rectangles frequently don't look very good on buttons
	 * having 3-D graphics; in any case, you can tell if
	 * the button has the input focus if the border around it is
	 * thick.  But you can use this code to draw a focus rectangle
	 * if you wish. */
#ifdef JKJ
	if (new_state & ODS_FOCUS)
		{
		lp_surface = surface_of(button, new_state & ODS_SELECTED);
		InflateRect(lp_surface,
			- WIDTH(lp_surface) / 8, - HEIGHT(lp_surface) / 8);
		DrawFocusRect(lpdi->hDC, lp_surface);
		} /* Button has focus */
#endif
	} /* paint_3d_button */
/* BTN_FNCS.H: Declarations of Support Functions for 3-D Buttons
   Include after WINDOWS.H

   Copyright (c) 1992 Springtime Software
   All Rights Reserved

*/

/* Fatal error procedure type */
typedef void ERRPROC(LPSTR msg, long info);

/* Initialize 3-D button facility (call from start of application) */
void	init_3d(LPSTR app_name, HANDLE instance, ERRPROC fatal);

/* Terminate 3-D button facility (call from end of application) */
void	end_3d(void);

/* Initialize or terminate all 3-D buttons in a dialog box */
/* (call from WM_INITDIALOG and WM_DESTROY) */
void	scan_3d_buttons(HWND dlg, LPARAM pass);

/* Paint a 3-D button in a given new state (call from WM_DRAWITEM) */
void	paint_3d_button(DWORD lParam);

/* For debugging: Display given DC+bitmap in center of screen */
void	show(HDC dc_in);

