// wuiobjdb.c - implement WUIMAN object database.
// Copyright (c) 1993 by Ron Burk

#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>      // sscanf()

#include "wuidbint.h"
#include "wuierror.h"
#include "wuistd.h"
#include "wuiobjdb.h"

// Hmm, these should probably be in wuiobjdb.h
static  const   TWuiName    M_DELETE("m_Delete");
static  const   TWuiName    M_ISA("m_IsA");
static  const   TWuiName    M_RENAME("m_Rename");
static  const   TWuiName    M_INITIALLOAD("m_InitialLoad");
static  const   TWuiName    P_CHILD("p_Child");

static  const   TWuiName    S_CLASSES("Classes");
static  const   TWuiName    S_ROOT("/");

////////////////////////////////////////////////////////////
// TWuiObjects - Provide arrays of AWuiObject pointers

// TWuiObjects - initialize object with desired array size.
TWuiObjects::TWuiObjects(int Size)
    :   TWuiGenericList(Size)
    {   MEMBERASSERT();
    }

// (private)
// Find - Return position of named child

int     TWuiObjects::Find(const TWuiName Which)
    {   MEMBERASSERT();
    for(int iChild = 0; iChild < NObjects(); ++iChild)
        if(Get(iChild)->Name() == Which)
            return iChild;
    return -1;
    }

void   *TWuiObjects::Insert(AWuiObject *NewObject, const TWuiName To)
    {   MEMBERASSERT();
    ASSERT(NewObject != 0);
    return Insert(NewObject, Find(To));
    }

void   *TWuiObjects::Insert(AWuiObject *NewObject, int Position)
    {   MEMBERASSERT();
    ASSERT(NewObject != 0);
    return TWuiGenericList::Insert((void *)NewObject, Position);
    }

AWuiObject  *TWuiObjects::Get(const TWuiName Which)
    {   MEMBERASSERT();
    int     Position = Find(Which);
    if(Position >= 0)
        return Get(Position);
    else
        return 0;
    }

AWuiObject *TWuiObjects::Remove(const TWuiName Which)
    {   MEMBERASSERT();
    int     Position = Find(Which);
    if(Position >= 0)
        return Remove(Position);
    else
        return 0;
    }

/////////////////////////////////////////////////
// AWuiObject - supply minimal object functions

AWuiObject     *AWuiObject::Root        = 0;
int             AWuiObject::TraceFlag   = 1;
TWuiDatabase   *AWuiObject::Database    = 0;

AWuiObject::AWuiObject(const TWuiName ObjectName,
                                      const char * /*Value*/)
    :   Name_(ObjectName)
    {   MEMBERASSERT();
    }

// WUIMAN Object destructor - delete objects in list.
AWuiObject::~AWuiObject()
    {   MEMBERASSERT();
    AWuiObject  *Object = 0;

    TWuiObjects *List = Children();
    if(List)
        {
        while((Object = List->Remove()) != NULL)
            delete Object;
        ASSERT(List->NObjects() == 0);
        delete List;
        }

    List    = Attributes();
    if(List)
        {
        while((Object = List->Remove()) != NULL)
            delete Object;
        ASSERT(List->NObjects() == 0);
        delete List;
        }
    }

void    AWuiObject::Trace(const char *Function,
                              TWuiPath Path, TWuiArgs *Args)
    {
    char    AttributeName[TWuiName::MAXNAME];
    char    MyName[TWuiName::MAXNAME];
    Args->AttributeName().GetName(AttributeName);
    this->Name().GetName(MyName);
    DebugPrintf("  [%s][%s] AbsPath = '%s', RelPath = '%s'"
                " Attribute='%s'",
                                            Function, MyName,
                    Path.AbsolutePath(), Path.RelativePath(),
                    AttributeName);
    }

// Get - default get attribute handler
//      NOTE: the default is to forward the message, or to return
//            FALSE if this object is the destination.
long    AWuiObject::Get(TWuiPath Path, TWuiArgs *Args,
                        AWuiObject *Attribute, char *Buffer,
                                           size_t MaxLength)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        Trace("AWuiObject::Get", Path, Args);
    long    Result  = TRUE; // assume success
    if(Path.Descend())      // if message is for a child
        {
        // then try to locate child and pass it message
        TWuiName ChildName      = Path.CurrentName();
        TWuiObjects *MyChildren = Children();
        AWuiObject  *Child      = NULL;
        if(MyChildren)
            Child   = MyChildren->Get(ChildName);
        if(Child)
            Result  = Child->Get(Path, Args, Attribute,
                                         Buffer, MaxLength);
        else    // else, no child by that name!
{
        TWuiName ChildName      = Path.CurrentName();
Child = MyChildren->Get(ChildName);
            Result  = WUIMAN_ERR_BAD_PATH;
}
        }
    else        // else message is for me
        {
        TWuiObjects *MyAttributes = Attributes();
        TWuiName AttributeName = Args->AttributeName();
        // p_child() returns # of children
        // p_child(n) returns name of nth child
        if(AttributeName == P_CHILD)
            {
            TWuiObjects *MyChildren = Children();
            AWuiObject  *Child = 0;
            if(Args->NArgs() <= 0)
                return MyChildren
                             ? MyChildren->NObjects() : 0;
            else if(Args->NArgs() == 1 && MyChildren)
                {
                if(Args->GetType(0) == TWuiArgs::LONG_ARG)
                    {
                    long    ArgNumber = Args->GetLong(0);
                    if(ArgNumber >= 0 && ArgNumber
                                 < MyChildren->NObjects())
                        {
                        Child = MyChildren
                                    ->Get(int(ArgNumber));
                        Child->Name().GetName(Buffer, MaxLength);
                        return 0;
                        }
                    }
                else // else attr arg is wrong type
                    return WUIMAN_ERR_BAD_ATTR_ARG;
                return WUIMAN_ERR_BAD_ATTR_ARG;
                }
            }
        if(MyAttributes)
            {
            AWuiObject  *Target = MyAttributes->Get(Args->AttributeName());
            if(Target)
                return Target->Get(Path, Args, Attribute, Buffer, MaxLength);
            }
        Result  = WUIMAN_ERR_BAD_ATTR;
        }
    return Result;
    }

// Set - default Set attribute handler
//      NOTE: the default is to forward the message, or to return
//            FALSE if this object is the destination
long    AWuiObject::Set(TWuiPath Path,
                      TWuiArgs *Args, AWuiObject *Attribute,
                                         const char *Buffer)
    {   MEMBERASSERT();
    long    Result  = TRUE; // assume success
    if(Path.Descend())   // if message is for a child
        {
        TWuiName ChildName      = Path.CurrentName();
        TWuiObjects *MyChildren = Children();
        AWuiObject  *Child      = NULL;
        if(MyChildren)
            Child   = MyChildren->Get(ChildName);
        if(Child)
            {
            Result  = Child->Set(Path, Args, Attribute,
                                                    Buffer);
            // if error or grandchild, just pass back result
            if(Result < WUIMAN_ERRORS || Path.Descend())
                ;
            // else, no error and message was for child
            else if(Args->AttributeName() == M_DELETE)
                {
                // Since tree has no upward pointers,
                // parent of target node must do deleting
                delete MyChildren->Remove(ChildName);
                
                if(MyChildren->NObjects() == 0)
                    {
                    }
                }
            }
        else    // else, no child by that name!
            Result  = WUIMAN_ERR_BAD_PATH;
        }
    else        // else message is for me
        {
        if(Args->AttributeName() == M_DELETE)
            ;   // parent will delete
        else if(Args->AttributeName() == M_ISA)
            return FALSE; // we ain't IsA nothin'!
        else if(Args->AttributeName() == M_RENAME)
            Name_   = Buffer;
        else if(Args->AttributeName() == M_INITIALLOAD)
            {   // create/load attributes and children
            ASSERT(Root != NULL);
            ASSERT(Root->Children() != NULL);
            AWuiObject  *Classes = Root->Children()->
                                             Get(S_CLASSES);
            ASSERT(Classes != NULL);
            ASSERT(Database != NULL);
            TWuiObjects *ChildList  = Children();
            TWuiObjects *AttrList   = Attributes();
            TWuiDatabaseKeys *Keys  = Database->
                              GetKeys(Path.AbsolutePath());
            const int MAX_ITEM = 1024;
            char *Buffer = new char[MAX_ITEM];
            const char *Name = Keys->FirstKey();
            while(Name != NULL)
                {
                TWuiObjects *List = ChildList;
                // Read class name, value string
                Database->ReadItem(Path.AbsolutePath(),
                                     Name, Buffer, MAX_ITEM);
                int IsAttribute = (*Name == '.');
                if(IsAttribute)
                    {
                    ++Name;      // skip over '.'
                    List    = AttrList;
                    }
                char *Value = strchr(Buffer, ',');
                if(Value)
                    *Value++    = '\0';
                AWuiObject *Master = Classes->Children()->Get(Buffer);
                if(Master)
                    {
                    AWuiObject *New = Master->New(Name, Value);
                    if(New)
                        {
                        List->Insert(New);
                        
                        }
                    }
                Name    = Keys->NextKey();
                }
            }
        else
            Result  = WUIMAN_ERR_BAD_ATTR;
        }
    return Result;
    }

TWuiObjects *AWuiObject::Children()
    {   MEMBERASSERT();
    // default is to not support children
    return NULL;
    }

TWuiObjects *AWuiObject::Attributes()
    {   MEMBERASSERT();
    // default is to not support attributes
    return NULL;
    }


////////////
// 11/06/93

static TWuiRegister RegisterContainer(
    new TWuiContainer("TContainer"));

// TWuiContainer - An AWuiObject that supports children and attributes.
TWuiContainer::TWuiContainer(const TWuiName ObjectName, const char *Value)
    :   AWuiObject(ObjectName)
    {   MEMBERASSERT();
    int     NChildren = 0, NAttributes = 0;
    if(Value)
        sscanf(Value, "%d,%d", &NChildren, &NAttributes);
    Children_   = new TWuiObjects(NChildren);
    Attributes_ = new TWuiObjects(NAttributes);
    }
AWuiObject  *TWuiContainer::Clone()
    {   MEMBERASSERT();
    char    InitialValue[32];
    sprintf(InitialValue, "%d,%d", Children()->NObjects(),
        Attributes()->NObjects());
    return new TWuiContainer(Name(), InitialValue);
    }
AWuiObject  *TWuiContainer::New(const TWuiName Name, const char *Value)
    {   MEMBERASSERT();
    return new TWuiContainer(Name, Value);
    }
TWuiObjects *TWuiContainer::Children()
    {   MEMBERASSERT();
    return Children_;
    }
TWuiObjects *TWuiContainer::Attributes()
    {   MEMBERASSERT();
    return Attributes_;
    }

// TWuiRegister - Register a new WUIMAN class.
TWuiRegister::TWuiRegister(AWuiObject *MasterCopy)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        {
        char Name[32];
        MasterCopy->Name().GetName(Name);
        DebugPrintf("TWuiRegister(\"%s\")", Name);
        }

    AWuiObject  *Classes = 0;
    if(AWuiObject::Root == NULL)    // if not initialized
        {
        // create root container object
        AWuiObject::Root    = new TWuiContainer(S_ROOT);
        // add "/Classes", to hold registered classes
        Classes             = new TWuiContainer(S_CLASSES);
        AWuiObject::Root->Children()->Insert(Classes);
        }
    ASSERT(AWuiObject::Root != NULL);
    ASSERT(AWuiObject::Root->Children() != NULL);
    Classes = AWuiObject::Root->Children()->Get(S_CLASSES);
    ASSERT(Classes != NULL);
    // Place object of class "X" in "/Classes/X".
    Classes->Children()->Insert(MasterCopy);
    }

// TWuiPath - Handle moving up and down the database "path"
TWuiPath::TWuiPath(char *TotalPath)
    :   AbsolutePath_(TotalPath), RelativePath_(TotalPath)
    {   MEMBERASSERT();
    ASSERT(TotalPath != NULL);
    ASSERT(TotalPath[0] == '/');
    }

// Descend() - Move relative path down one node.
int TWuiPath::Descend()
    {
    char    *Child = strchr(RelativePath_, '/');
    if(!Child)   // if no children
        return FALSE;
    // ignore trailers (e.g. "/a/b/c/")
    else if(Child[1] == '\0')
        return FALSE;
    // else there's *some* kind of child name
    else
        {
        RelativePath_   = Child+1;
        return TRUE;
        }
    }

// Ascend() - Move relative path up one node.
int TWuiPath::Ascend()
    {
    switch(RelativePath_ - AbsolutePath_)
        {
        case    0:
            return FALSE;
        case    1:
            RelativePath_   = AbsolutePath_;
            return TRUE;
        default:
            RelativePath_   -= 2;
            while(RelativePath_ > AbsolutePath_)
                if(*RelativePath_ == '/')
                    break;
                else
                    --RelativePath_;
            ASSERT(*RelativePath_ == '/');
            ++RelativePath_;
            return TRUE;
        }
    }

// CurrentName() - return first name of relative path.
TWuiName    TWuiPath::CurrentName()
    {   MEMBERASSERT();
    char        Name[TWuiName::MAXNAME+1];
    const char *Rover = strchr(RelativePath_, '/');
    int         Length;
    if(Rover)
        Length  = Rover-RelativePath_;
    else
        {
        Length  = strlen(RelativePath_);
        Rover   = RelativePath_;
        }
    strncpy(Name, RelativePath_, Length);
    Name[Length]    = '\0';
    return TWuiName(Name);
    }

// Push() - Append to the path.
int     TWuiPath::Push(const char *SubPath)
    {   MEMBERASSERT();
    char   *Rover = AbsolutePath_ + strlen(AbsolutePath_);
    if(Rover > AbsolutePath_)
        --Rover;    // point to char before EOS
    if(*Rover != '/')
        strcat(AbsolutePath_, "/");
    strcat(AbsolutePath_, SubPath);
    return TRUE;    // Need to check for overflow!
    }

// Pop() - Pop the last node off the path.
int TWuiPath::Pop()
    {   MEMBERASSERT();
    char   *Rover = AbsolutePath_ + strlen(AbsolutePath_);
    if(Rover > AbsolutePath_)
        --Rover;    // point to char before EOS
    if(Rover > AbsolutePath_ && *Rover == '/')
        --Rover;    // skip any bogus trailing '/'
    while(Rover > AbsolutePath_ && *Rover != '/')
        --Rover;
    if(Rover <= AbsolutePath_)
        return FALSE;
    else    // else Rover points at '/'
        {
        *Rover  = '\0';
        return TRUE;
        }
    }

// TWuiAttribute - Base class for attributes.
TWuiAttribute::TWuiAttribute(const TWuiName ObjectName, const char *Value)
    :   AWuiObject(ObjectName)
    {   MEMBERASSERT();
    if(Value)
        Value_  = StringClone(Value);
    else
        Value_  = 0;
    }
AWuiObject  *TWuiAttribute::Clone()
    {   MEMBERASSERT();
    return new TWuiAttribute(Name(), Value_);
    }
AWuiObject  *TWuiAttribute::New(const TWuiName Name, const char *Value)
    {   MEMBERASSERT();
    return new TWuiAttribute(Name, Value);
    }
long    TWuiAttribute::Get(TWuiPath Path, TWuiArgs *Args,
      AWuiObject *Attribute, char *Buffer, size_t MaxLength)
    {
    if(AWuiObject::TraceFlag)
        Trace("TWuiAttribute::Get", Path, Args);
    if(Value_ == NULL)
        {
        // then return inherited value if any
        if(Attribute)
            return Attribute->Get(Path, Args, NULL, Buffer,
                                                 MaxLength);
        else    // else caller gets a NULL notification
            {
            if(Buffer)
                *Buffer = '\0';
            return WUIMAN_ERR_NULL_ATTR;
            }
        }
    else    // else try to copy value to caller's buffer
        {
        size_t  ValueLength = strlen(Value_) + 1;
        long    Status = ValueLength;
        if(ValueLength > MaxLength)
            {
            ValueLength = MaxLength;
            Status      = WUIMAN_ERR_ATTR_TOO_LONG;
            }
        memcpy(Buffer, Value_, ValueLength);
        return Status;
        }
    }
long    TWuiAttribute::Set(TWuiPath /*Path*/,
            TWuiArgs * /*Args*/, AWuiObject * /*Attribute*/,
                                         const char *Buffer)
    {
    if(Value_)
        {
        delete[] Value_;
        Value_  = 0;
        }
    if(Buffer)
        Value_  = StringClone(Buffer);
    return 0;
    }


// TWuiPersistantAttribute - Base class for attributes.
TWuiPersistantAttribute::TWuiPersistantAttribute(
               const TWuiName ObjectName, const char *Value)
    :   TWuiAttribute(ObjectName, Value)
    {   MEMBERASSERT();
    }
AWuiObject  *TWuiPersistantAttribute::Clone()
    {   MEMBERASSERT();
    return new TWuiPersistantAttribute(Name(), Value());
    }
AWuiObject  *TWuiPersistantAttribute::New(
                     const TWuiName Name, const char *Value)
    {   MEMBERASSERT();
    return new TWuiPersistantAttribute(Name, Value);
    }
// Here's where we implement persistance
long    TWuiPersistantAttribute::Set(TWuiPath Path,
                      TWuiArgs *Args, AWuiObject *Attribute,
                                         const char *Buffer)
    {
    long    Status = TWuiAttribute::Set(Path, Args,
                                         Attribute, Buffer);
    Database->WriteItem(Path.AbsolutePath(),
                            Args->AttributeName(), Value());
    return Status;
    }

