/*////////////////////////////////////////////////////////////////////////////

    MODULE:     PICEDIT.H
    
    Description:
        This is the declaration of the CPicEdit class.  This class will
        provide masked editing for any CEdit.  The user of this class can
        pass in any edit mask and the control will be subclassed to handle
        all keyboard and mouse input.
        

        The following is a list of the valid mask characters:
        
        Mask Character  Description
        ----------------------------------------------------------------------
        9               Allows only numbers
        A               Allows only alphabetic letters (a through z and A - Z)
        N               Allows numbers or alpha letters
        L               Allows only letters and foces it to lower case
        U               Allows only letters and forces it to upper case
        X               Allows any ASCII character
        /               Escape character (treat next character as literal)
        
        To set a phone number mask, you would use "(999) 999-9999" as your
        mask string (See Also the predefined masks described below).  To
        make a mask that forced the third character to be an upper case
        letter and all others to be numeric, your mask would be "99U999999".
        
        The class can be used for an edit box on a CDialog or a CFormView
        class.  You must call the SubclassEdit() method before setting the
        mask string.
        
    History:                           
        Date        Email       Action
        ----------------------------------------------
        08/17/93    daveste     created
                
*/////////////////////////////////////////////////////////////////////////////

#ifndef __PICEDIT_H__
#define __PICEDIT_H__

#undef AFXAPP_DATA
#define AFXAPP_DATA AFXAPI_DATA

/*////////////////////////////////////////////////////////////////////////////
    The following defines are for the predefined edit masks that come with
    this class.  These can be used in the SetMask() method that takes an
    int ID instead of a literal string.  Using these will give your decendant
    classes a common place of reference for standard masks, and will ease the
    pain in modifications when the value of these predefined masks must change
    (since everything will eventually change!).    

    ID                  Mask                    Description
    ---------------------------------------------------------------------------
    PEM_US_PHONE        (999) 999-9999          Standard US/Canada phone number
    PEM_US_PHONE_EX4    (999) 999-9999 [9999]   Phone with 4 digit extension
    PEM_US_PHONE_EX5    (999) 999-9999 [99999]  Phone with 5 digit extension
    PEM_SSN             999-99-9999             Social Security Number
    PEM_US_ZIP          99999                   US ZIP code (5 digit only)
    PEM_US_ZIP_4        99999-9999              US ZIP + 4 code
    PEM_US_ZIP_6        99999-999999            US ZIP + 6 code
*/////////////////////////////////////////////////////////////////////////////
#define PEM_US_PHONE        1
#define PEM_US_PHONE_EX4    2
#define PEM_US_PHONE_EX5    3
#define PEM_SSN             4
#define PEM_US_ZIP          5
#define PEM_US_ZIP_4        6
#define PEM_US_ZIP_6        7

//This is the amount that data mask characters will be increased by in order
//to distinguish them from literal mask characters
#define INFLATION_FACTOR    128

//forward declaration
class CPicEdit;

//////////////////////////////////////////////////////////////////////////////
// METHODS FOR USE IN A DoDataExchange BLOCK

//the DDX_PicEditControl can be used just like DDX_Contol to associate a window
//with a CPicEdit object.  It will call the SubclassEdit() member function and
//the SetMask() member function
void AFXAPI DDX_PicEditControl(CDataExchange* pDX, int nIDC, CPicEdit& rPicEdit,
								int iMaskID);
void AFXAPI DDX_PicEditControl(CDataExchange* pDX, int nIDC, CPicEdit& rPicEdit,
								LPCSTR lpszMask);

//The DDX_PicEditText can be used to set and get data from the CPicEdit control.
//This will call GetData() and SetData() depending on the direction of data exchange.
void AFXAPI DDX_PicEditText(CDataExchange* pDX, CPicEdit& rPicEdit, CString& value,
							BOOL fTrim = FALSE);

//////////////////////////////////////////////////////////////////////////////
// CPICEDIT CLASS 

class CPicEdit : public CEdit
{ 
    //PUBLIC INTERFACE
    public:     
        //ctor, dtor
        CPicEdit();
        virtual ~CPicEdit();
    
        //starts subclassing of the control
        BOOL SubclassEdit(UINT iCtlID, CWnd* pParentWnd);
        BOOL SubclassEdit(HWND hEdit);
        
        //sets the input mask for the control
        BOOL SetMask(LPCSTR rstrMask);
        //set the input mask for the control from a list of
        //predefined input masks (see method for list of masks)
        BOOL SetMask(int iMaskID);
        
        //sets the prompt character
        void SetPromptChar(char chNewPrompt);
        //gets the prompt character
        char GetPromptChar();
        
        //sets the data in the edit
        BOOL SetData(const CString& rstrData, int iStartPos = 0, BOOL fInsert = FALSE,
        				BOOL fIgnorePromptChars = TRUE);
        BOOL SetData(LPCSTR lpszData, int iStartPos = 0, BOOL fInsert = FALSE,
        				BOOL fIgnorePromptChars = TRUE);
        
        //returns just the data in the control (no mask chars)
        void GetData(LPSTR lpszData, int iBufferLen, BOOL fTrim = FALSE);
        void GetData(CString& sData, BOOL fTrim = FALSE);        
        
        //clears all data from the control
        void ClearData(void);
        
        //returns the entire contents of the edit (including mask chars)
        CString& GetString(void);
        
        //sets the auto advance property
        void SetAutoAdvance(BOOL fAutoAdvance);
        //set the auto select property
        void SetAutoSelect(BOOL fAutoSelect);
        //set insert mode capabilities on or off
        void SetInsertModeCapable(BOOL fInsertModeCapable);
        //set insert mode on or off
        void SetInsertMode(BOOL fInsertMode);
        //get the current state of the insert mode
        BOOL GetInsertMode();
        //set the read-only property
        void SetReadOnly(BOOL fReadOnly = TRUE);
        //get the current status of the read-only property
        BOOL GetReadOnly();
    
    //{{AFX_MSG(CPicEdit)
    afx_msg void OnSetFocus(CWnd* pOldWnd);
    afx_msg void OnKillFocus(CWnd* pNewWnd);
    afx_msg void OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags);
    afx_msg void OnChar(UINT nChar, UINT nRepCnt, UINT nFlags);
    //afx_msg int OnMouseActivate(CWnd* pDesktopWnd, UINT nHitTest, UINT message);
    afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
    afx_msg void OnMouseMove(UINT nFlags, CPoint point);
    afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
    virtual afx_msg LRESULT OnEditCopy(WPARAM, LPARAM);
    virtual afx_msg LRESULT OnEditPaste(WPARAM, LPARAM);
    virtual afx_msg LRESULT OnEditCut(WPARAM, LPARAM);
    virtual afx_msg LRESULT OnEditUndo(WPARAM, LPARAM);
    virtual afx_msg LRESULT OnEditCanUndo(WPARAM, LPARAM);
    virtual afx_msg LRESULT OnEditEmptyUndoBuffer(WPARAM, LPARAM);
    virtual afx_msg LRESULT OnClear(WPARAM, LPARAM);
    afx_msg void OnPaint();
    //}}AFX_MSG
    
        
    //PROTECTED MEMBER FUNCTIONS
    protected:
        //Validate a character at a given position
        virtual BOOL IsCharValid(UINT nChar, char& ch, const CString& rstrMask);
        //Refresh the text of the control
        virtual void BuildDisplayStr(CString& rstrMask, CString& rstrDisplay);
        //parse a mask string into our interal format
        virtual void ParseMaskString(CString& rstrMask);
        //is the mask character a literal
        virtual BOOL IsMaskLiteral(char ch);
        //refresh the control's text
        virtual void RefreshDisplay(void);
        //set the cursor position forward or backward
        virtual int SetCursorPos(int nOldPos, int iDirection, BOOL fSelect = FALSE);
        //find the next valid data position
        virtual int GetNextDataPos(int nOldPos);
        //find the previous valid data position
        virtual int GetPrevDataPos(int nOldPos);
		//finds the last filled data pos based on a prev pos
		virtual int FindLastFilledDataPos(int nStartPos);
        //clears the data from a range
        virtual int ClearDataRange(int iStart, int iEnd);
        //push data characters down one position (used for insert mode)
        virtual BOOL PushChar(int iStartPos, char ch, BOOL fRefreshDisplay = TRUE);
        //pull data characters back one (used for insert mode);
        virtual void PullChar(int iFromPos, int iToPos, BOOL fSetUndoBuffer = TRUE);
        //advance the focus to the next control in the tab order
        virtual void AdvanceFocusNext();
        //advance the focus to the previous control in the tab order
        virtual void AdvanceFocusPrevious();
        //get the repaint position for optimization
        virtual int GetRepaintPos(int iIndex);
        
    //PROTECTED DATA
    protected:
        CString m_strMask;      //edit mask
        CString m_strDisplay;   //edit display string
        char m_chPrompt;        //prompt character
        BOOL m_fSubclassed;     //are we subclassed?
        CString m_strMaskChars; //list of mask charactrs
        char m_chLiteralChar;   //the literal char indicator
        BOOL m_fGotKey;         //flag for telling OnChar that OnKeyDown
                                //already responded.
        BOOL m_fMouseActivate;  //true if control is activated by mouse (don't select!)                        
        BOOL m_fAutoAdvance;    //if true, focus will advance to next control
                                //after entry of last data character 
        BOOL m_fAutoSelect;     //if true, entire text will be selected when
                                //the control gets focus.
        BOOL m_fInsertModeCapable; //if the user does not want this control
                                   //to be insert mode capable, they can set
                                   //this property to false
        BOOL m_fInsertMode;     //in insert mode, characters are pushed and
                                //pulled like a normal edit box--otherwise it
                                //operates in overtype mode.
		BOOL m_fReadOnly;		//flag indicating read-only status                            
        CString m_sUndoBuffer;  //undo buffer for control    
        UINT m_iLastKey;        //the code for the last key pressed in the control
        int m_nRepaintPos;      //the point at which to paint from.  This indicates
                                   //the leftmost point that needs to be painted.
		BOOL m_fEatMouseMove;	//flag to eat mouse move messages
		CPoint m_ptMouseMoveRef;//the point of the current cursor so that we can ignore
								  //mouse moves if they are to the right of the last
		        
    DECLARE_MESSAGE_MAP()
    DECLARE_DYNAMIC(CPicEdit)
};

///////////////////////////////////////////////////////////////////////////////
// INLINE FUNCTIONS

inline CString& CPicEdit::GetString()
{
    return m_strDisplay;
} // GetEditString()
                       
inline BOOL CPicEdit::IsMaskLiteral(char ch)
{
    return ((unsigned char)ch < INFLATION_FACTOR);
}   

inline void CPicEdit::RefreshDisplay(void)
{
    SetWindowText(m_strDisplay);
    if(!m_fReadOnly)
    	SetModify(TRUE);
}

inline void CPicEdit::SetAutoAdvance(BOOL fAutoAdvance)
{
    m_fAutoAdvance = fAutoAdvance;
}
                                    
inline void CPicEdit::SetAutoSelect(BOOL fAutoSelect)
{
    m_fAutoSelect = fAutoSelect;
}

inline char CPicEdit::GetPromptChar()
{
    return m_chPrompt;
}

inline void CPicEdit::SetInsertModeCapable(BOOL fInsertModeCapable)
{
    m_fInsertModeCapable = fInsertModeCapable;
}       

inline void CPicEdit::SetInsertMode(BOOL fInsertMode)
{
    m_fInsertMode = fInsertMode;
}       

inline BOOL CPicEdit::GetInsertMode()
{
    return m_fInsertMode;
}

inline void CPicEdit::AdvanceFocusNext()
{
    ((CDialog*)GetParent())->NextDlgCtrl();
}

inline void CPicEdit::AdvanceFocusPrevious()
{
    ((CDialog*)GetParent())->PrevDlgCtrl();
}

inline void CPicEdit::SetReadOnly(BOOL fReadOnly)
{
	m_fReadOnly = fReadOnly;
} 

inline BOOL CPicEdit::GetReadOnly()
{
	return m_fReadOnly;
}

#undef AFXAPP_DATA
#define AFXAPP_DATA NEAR

#endif // __PICEDIT_H__