#include "stdafx.h"         
#include "io.h"
#include "bmp_samp.h"

BOOL FAR PASCAL CreateBitmapFromFile(LPSTR lpstrFile, HBITMAP *hBitmap, HPALETTE *hPal)
//lpstrFile = pointer to name of bitmap file
//hBitmap = bitmap handle to return new bitmap in
//hPal = palette handle for new palette (only needed for 256 color bitmaps)
{
  int i, fp;				//loop counter and file handle
  int numEntries;			//number of palette entries in bitmap
  HGLOBAL hGlb;				//global handle used to read in bitmap data
  LPBYTE bmpData;			//pointer to global memory used to hold bitmap data
  LPBYTE dataBits;			//pointer to actual bitmap picture data
  BITMAPFILEHEADER bmfh;		//bitmap file header
  LPBITMAPINFOHEADER bmih;		//pointer to bitmap info header
  HDC hDC;				//device context used to create hBitmap
  LOGPALETTE *LogPal=NULL;		//memory pointer to store logical palette in
  LPRGBQUAD lpRGB;			//pointer to RGB quad structures of bitmap        
  OFSTRUCT datafile;

  if (!hBitmap) {
    //ERROR: invalid parameter
    return FALSE;
  }

  //open bitmap file
  fp = OpenFile(lpstrFile, (LPOFSTRUCT)&datafile, OF_READ);
  if (fp == -1) {
    //ERROR: unable to open file
    return FALSE;
  }
  
  //read in bitmap file header
  _lread(fp, (LPSTR)&bmfh, sizeof(BITMAPFILEHEADER));

  //verify file is type BM (bitmap file)
  if (bmfh.bfType != 0x4d42) {
    //ERROR: invalid file
    _lclose(fp);
    return FALSE;
  }

  //allocate global memory to store bitmap data in
  hGlb = GlobalAlloc(GHND, _filelength(fp));
  bmpData = (LPBYTE)GlobalLock(hGlb);

  //verify memory properly allocated
  if (!bmpData) {
    if (hGlb) GlobalFree(hGlb);
    _lclose(fp);
    //ERROR: memory allocation failed
    return FALSE;
  }

  //read in bitmap
  _hread(fp, (LPSTR)bmpData, bmfh.bfSize-sizeof(BITMAPFILEHEADER));

  _lclose(fp);	//no longer need file handle

  //get pointer to bitmap info header
  bmih = (LPBITMAPINFOHEADER)bmpData;

  //get pointer to actual bitmap data
  dataBits = (LPBYTE)((LONG)bmpData + (LONG)bmfh.bfOffBits - sizeof(BITMAPFILEHEADER));

  //get pointer to RGB quad structures
  lpRGB = (LPRGBQUAD)((LONG)bmih + sizeof(BITMAPINFOHEADER));

  //determine the number of palette entries based on number of colors
  if (bmih->biBitCount == 1) numEntries = 2;
  else if (bmih->biBitCount == 4) numEntries = 16;
  else if (bmih->biBitCount == 8) numEntries = 256;
  else if (bmih->biBitCount == 24) numEntries = 0;		//no palette for 24 bit color

  //get a screen device context
  hDC = GetDC(NULL);

  //read in logical palette
  if (numEntries && hPal) {
    //get memory to read palette into
    LogPal = (LOGPALETTE*)LocalAlloc(LPTR, sizeof(LOGPALETTE)+numEntries*sizeof(PALETTEENTRY));

    //setup logpalette structure
    LogPal->palVersion = 0x300;
    LogPal->palNumEntries = numEntries;

    for (i=0;i < numEntries;i++) {
      LogPal->palPalEntry[i].peRed = lpRGB[i].rgbRed;
      LogPal->palPalEntry[i].peGreen = lpRGB[i].rgbGreen;
      LogPal->palPalEntry[i].peBlue = lpRGB[i].rgbBlue;
      LogPal->palPalEntry[i].peFlags = 0;
    }

    //create the palette using the bitmap's logical palette
    *hPal = CreatePalette((LPLOGPALETTE)LogPal);

    //set palette into the device context
    SelectPalette(hDC, *hPal, TRUE);
    RealizePalette(hDC);  
  }
  else if (hPal) *hPal = NULL;

  //create the bitmap in the device context
  *hBitmap =  
          CreateDIBitmap(hDC,bmih,CBM_INIT,(LPSTR)dataBits,(LPBITMAPINFO)bmih,DIB_RGB_COLORS);

  //perform cleanup
  ReleaseDC(NULL, hDC);

  if (LogPal) LocalFree(LogPal);

  GlobalUnlock(hGlb);
  GlobalFree(hGlb);

  //verify bitmap created
  if (!*hBitmap) {
    if (hPal && *hPal) DeleteObject(*hPal);
    //ERROR: create bitmap failure
    return FALSE;
  }

  return TRUE;
}

BOOL FAR PASCAL DrawBitmap(HBITMAP hBitmap, HPALETTE hPal, HDC hDC, int x, int y)
//hBitmap = handle of bitmap to be drawn
//hPal = palette of bitmap if necessary
//hDC = device context bitmap to be drawn on
//x, y = location bitmap to be drawn on device context
{
  BITMAP bitmap;		//structure to hold bitmap statistics
  HDC    hDCMem;		//memory DC to place bitmap in
  POINT  bSize, origin;

  if (!hBitmap || !hDC) {
    //ERROR: invalid parameter
    return FALSE;
  }

  //create a compatible memory dc to place bitmap in
  hDCMem = CreateCompatibleDC(hDC);

  //select bitmap into memory dc
  SelectObject(hDCMem, hBitmap);

  //set mapping modes to be equivalent
  SetMapMode(hDCMem, GetMapMode(hDC));

  //get statistics of bitmap
  GetObject(hBitmap, sizeof(BITMAP), (LPSTR)&bitmap);

  //convert size and origin of bitmap to logical coordinates
  bSize.x = bitmap.bmWidth;
  bSize.y = bitmap.bmHeight;
  DPtoLP(hDC, &bSize,1);

  origin.x = origin.y = 0;
  DPtoLP(hDCMem,&origin,1);

  //if a palette is provided, set dc to use it
  if (hPal) {
    SelectPalette(hDC, hPal, FALSE);
    RealizePalette(hDC);
  }

  //draw bitmap onto passed in device context
  BitBlt(hDC, x, y, bSize.x, bSize.y, hDCMem, origin.x, origin.y, SRCCOPY);

  //cleanup
  DeleteDC(hDCMem);

  return TRUE;
}

