;    TIME_FNS.ASM  -  time processes in clock pulse units
;    Mike Dawdy
;    76275.700@compuserve.com
;    http://ourworld.compuserve.com/homepages/Mike_Dawdy

;    The clock rate is 1.1931818 mega-ticks per second
;    One clock cycle is 838.09652 nanoseconds

     .model    small

;    variables used for analysis of results
     public    _pit_latch;PIT latch register
     public    _pic_irr  ;PIC IRR register

NORMAL    equ  1         ;conditional assembly identifier

;    8253/4 programmable interval timer (PIT)
PIT_CWR   equ  43h       ;control word register
PIT_CTR0  equ  40h       ;counter-0 I/O register

;    8259A programmable interrupt controller (PIC)
PIC_A0    equ  20h       ;PIC A0 address line
PIC_A1    equ  21h       ;PIC A1 address line
PIC_EOI   equ  00100000b ;non-specific end of interrupt
PIC_SEOI  equ  01100000b ;specific EOI for IRQ zero

     .data

timer_vec dd   ?         ;save DOS timer ISR segment:offset
isr_time  dw   ?         ;time for interrupt service routin
pic_imr   db   ?         ;PIC interrupt mask register
_pic_irr  db   ?         ;PIC interrupt request register
          even
_pit_latch dw  ?         ;PIT latch register
latch     dw   ?         ;PIT latch register, local calcs

     .code

_init_time proc
     public  _init_time          ;unsigned init_time(void);

;    1. save timer int vector, point it to our timer_isr()

     mov     al,08h      ;this is restored by reset_time()
     mov     ah,35h
     int     21h
     mov     word ptr timer_vec+2,es
     mov     word ptr timer_vec,bx

     mov     dx,offset timer_isr
     mov     ax,cs
     push    ds
     mov     ds,ax
     mov     al,08h
     mov     ah,25h
     int     21h
     pop     ds

;    2a. get time for 8253/4 interrupt routine timer_isr()
;    for use in calculating net elapsed time in read_time()

     cli
     in      al,PIC_A1   ;save DOS 8259A IMR value
     mov     pic_imr,al
     mov     al,11111110b;mask all ints except PIT
     out     PIC_A1,al

     mov     word ptr cs:Timer+2,0
     mov     word ptr cs:Timer,0

     mov     al,00010100b;counter-0,lsb,mode 2,bin
     out     PIT_CWR,al  ;write the control word
     jmp     short $+2   ;recovery time for I/O chip
     sub     al,al       ;write the initial count ..
     out     PIT_CTR0,al
     sti                 ;prepare to latch counter-0
     sub     al,al       ;al = 0000xxxx  = chopmodb
     hlt                 ;wait for clock interrupt

;    2b. the remaining count latched next, then 2s
;    complimented, indicates the time for timer_isr()

     out     PIT_CWR,al  ;move counter-0 to latch register
     cli
     jmp     short $+2   ;recovery time for I/O chip
     in      al,PIT_CTR0 ;read low byte from latch into ax
     neg     al
     sub     ah,ah
     mov     isr_time,ax ; = time for timer_isr()

;    3. initialization is done

     mov     al,pic_imr  ;restore DOS IMR
     out     PIC_A1,al
     mov     ax,isr_time ;establish return value
     sti

;    If the 8253/4 counter-0 runs out before start_time()
;    is called, timer_isr() will be entered and Timer will
;    be incremented again. This won't matter as
;    start_time() resets Timer to zero on each call.

     ret

_init_time endp

Timer dd     ?

timer_isr    proc far  ;void interrupt far timer_isr(void);

     inc     word ptr cs:Timer;1.increment doubleword Timer
     jnz     timer_exit
     inc     word ptr cs:Timer+2

timer_exit:
     push    ax               ;2. write EOI command to PIC
     mov     al,PIC_EOI
     out     PIC_A0,al
     pop     ax

     iret
timer_isr   endp

_start_time proc
     public  _start_time            ;void start_time(void);

;    An 8253/4 counter is programmed by writing a control
;    word and an initial count. The initial count 0 used
;    here is treated as 65536. After writing the control
;    word and initial count, the  counter will be loaded
;    in the next CLK pulse (within 838.09652 nsecs). OUT
;    goes low (an interrupt will occur) N (65536) pulses
;    after the initial count is written

;    On the other hand writing a new count while counting
;    (without writing a new control word) does not affect
;    the current counting sequence. The new count would not
;    be loaded until the end of the current counting cycle.

;    Writing a new control word will stop counting only on
;    some strict PC compatible machines; not PS/2s

;    1. write the control word and initial count

     mov     al,00110100b;counter-0,lsb/msb,mode 2,bin
     cli
     out     PIT_CWR,al  ;write the control word
     jmp     short $+2   ;recovery time for I/O chip
     sub     al,al       ;write the initial count
     out     PIT_CTR0,al ; .. lsb
     jmp     short $+2   ;recovery time for I/O chip
     out     PIT_CTR0,al ; .. msb

;    If counter-0 counts down to 1 any time after the
;    begining of the above cli and before the new count
;    is loaded, the pending 8253/4 interrupt will NOT be
;    cleared and we would we get an unwanted interrupt
;    after the sti at the end of this routine! This
;    would increment Timer and the result time would
;    be 64K clock cycles too large


;    2. We now need to clear any unwanted interrupt which
;    may occur as outlined above. After a poll command,
;    the 8259A PIC treats its next read pulse as an
;    interrupt acknowledge, sets the appropriate ISR bit
;    if there is a request and resets the corresponding
;    IRR bit. A specific EOI command can be used to reset
;    the ISR bit. This sequence will prevent a pending
;    interrupt from occuring.

IF NORMAL

     in      al,PIC_A0   ;read PIC IRR
     test    al,00000001b;is timer interrupt pending?
     jz      start_time_exit ;no

     mov     al,00001110b;poll command
     out     PIC_A0,al
     jmp     short $+2   ;recovery time for I/O chip
     in      al,PIC_A0   ;read command,ignore the data
                         ;al = i0000ppp
;    the data enabled onto the bus by this read indicates
;    if there is an interrupt and what the priority level
;    is of that interrupt. It should be noted that this
;    interrupt may not yet be posted in the IRR. All of
;    this information we do not need in this application.
;    We care only that this read is treated by the 8259A
;    as an interrupt acknowledge.

     jmp     short $+2   ;recovery time for I/O chip
     mov     al,PIC_SEOI ;specific EOI to reset IRR bit 0
     out     PIC_A0,al

;    3. start_time() is finished

start_time_exit:

ENDIF

     mov     word ptr cs:Timer,0
     mov     word ptr cs:Timer+2,0
     sti
     ret

_start_time endp

_read_time proc
     public _read_time      ;unsigned long read_time(void);

;    returns dx=ax=0 if the time interval overflows into
;    the high word of the doubleword timer interrupt counter.
;    This indicates that the time interval exceeded
;    approximately 1 hour.

;    A counter can be read "on the fly" without affecting
;    counting in progress using the counter latch command.
;    This command is written to the control word register
;    and the content of the counting element is transferred
;    to the output latch. The count is held in the latch
;    until it is read by the cpu or until the  counter is
;    reprogrammed (by writing a new control word).

;    1.  latch and read the counter

     sub     al,al       ;ax = 0000xxxx  = chopmodb
     cli                 ;   = counter-0 latch command
     out     PIT_CWR,al  ;move counter-0 to latch
     jmp     short $+2   ;recovery time for I/O chip
     in      al,PIT_CTR0 ;read low byte from latch
     jmp     short $+2   ;recovery time for I/O chip
     mov     ah,al
     in      al,PIT_CTR0 ;read high byte from latch
     xchg    ah,al
     neg     ax
     mov     _pit_latch,ax
     mov     latch,ax    ;for local calculations

;    Interrupts are serviced by the CPU only between
;    instruction executions. The counter can count down
;    to zero at any time.

;    If counter-0 were to count down to 1 during the
;    execution of the cli instruction or during the 1st
;    part of the following out instruction, it would
;    reload the initial count which would then be read by
;    the out instruction without an interrupt having been
;    generated to increment Timer, and the time
;    calculated next would be 64K clock cycles too small.

;    If that has happened, then BOTH the appropriate bit
;    (bit-0) will have been set in the PIC interrupt
;    request register(IRR) AND the value read from the
;    counter latch register will be large (close to the
;    initial count)

;    2. compute isr_time * Timer before incrementing Timer

     mov     ax,word ptr cs:Timer
     mul     isr_time

IF NORMAL

     push    ax          ;we need dx:ax in step 4 below

;    3. read the IRR, increment Timer if indicated

     in      al,PIC_A0   ;read PIC IRR
     mov     _pic_irr,al
     test    al,00000001b;was interrupt pending?
     jz      not_pending ;no
     cmp     latch,10    ;is latch close?
     ja      not_pending ;no

;    if an interrupt was pending and if latch is close to
;    initial count then counter-0 counted down to 1 as
;    outlined above and Timer is numerically 1 too small

;    the largest observed value for latch on a 4.77 mhz
;    PC was initial count-4. Initial count-10 gives a large
;    safety margin without being too large

     inc     word ptr cs:Timer
     jnz     not_pending
     inc     word ptr cs:Timer+2

not_pending:
     cmp     word ptr cs:Timer+2,0
     je      time_ok     ;high word != 0 => error
     sub     dx,dx       ;would simply require
     sub     ax,ax       ;triple-word arithmetic,
     mov     word ptr cs:Timer+2,ax
     jmp     error_exit  ;but let others do it

;    4. net time = Timer:latch - isr_time * Timer

time_ok:
     pop     ax

ENDIF

     sub     latch,ax
     sbb     word ptr cs:Timer,dx

;    5. reset PIT counter-0 to zero for next cycle

     mov     al,00110100b;counter-0,lsb/msb,mode 2,bin
     out     PIT_CWR,al  ;write the control word
     jmp     short $+2   ;recovery time for I/O chip
     sub     al,al ;write the initial count ..
     out     PIT_CTR0,al ; .. lsb
     jmp     short $+2   ;recovery time for I/O chip
     out     PIT_CTR0,al ; .. msb

;    6. remove any pending timer interrupt which may have
;    been posted to the IRR either as outlined in step 1,
;    or any time since the out instruction of step 1

IF NORMAL

     in      al,PIC_A0   ;read PIC IRR
     test    al,00000001b;is timer interrupt pending?
     jz      read_time_exit

     mov     al,00001110b;poll command
     out     PIC_A0,al
     jmp     short $+2   ;recovery time for I/O chip
     in      al,PIC_A0   ;read command,ignore the data
     jmp     short $+2   ;recovery time for I/O chip
     mov     al,PIC_SEOI ;specific EOI for level zero
     out     PIC_A0,al

read_time_exit:

ENDIF

     mov     dx,word ptr cs:Timer
     mov     ax,latch    ;establish long return value

error_exit:
     mov     word ptr cs:Timer,0
     sti                 ;must be done after step 6 above
     ret                 ;return with Timer:latch in dx:ax

_read_time endp

_reset_time proc
     public  _reset_time            ;void reset_time(void);

;    1. reset counter-0 to BIOS specs: mode 3 not mode 2

     cli
     mov     al,00110110b;counter-0,lsb/msb,mode3,bin
     out     PIT_CWR,al
     jmp     short $+2   ;recovery time for I/O chip
     sub     al,al       ;count for 18.2 pulses/sec
     out     PIT_CTR0,al
     jmp     short $+2   ;recovery time for I/O chip
     out     PIT_CTR0,al

;    2. restore DOS timer int vector

     mov     ax,word ptr timer_vec+2
     mov     dx,word ptr timer_vec
     push    ds
     mov     ds,ax
     mov     al,08h
     mov     ah,25h
     int     21h
     pop     ds

     sti
     ret

_reset_time endp

       end

