// bcc32 -w -DSTRICT strip.c
//      - OR -
// cl /DSTRICT /W3 strip.c
//
// Strips ".reloc" sections from files.
//
#include <stdio.h>
#include <stdlib.h>
#include <windows.h>

#define FAIL(Msg)   fprintf(stderr, "Can't strip: %s\n", Msg); return;

void DirectoryLoop(const char *dir);
void FileLoop(const char *dir);
void ProcessFile(const char *FullPathName);

// For recursive routine DirectoryLoop
static char SubDirName[MAX_PATH];

// Keeps track of space saved
static DWORD TotalSpaceSaved;

// Decides if "for keeps"
static BOOL ShowOnly;

int main(int argc, char **argv) {
  char *NamePtr;
  int ArgCount;

  TotalSpaceSaved = 0;
  ShowOnly = FALSE; // This could be set via the command line

  if (argc == 1) {
    printf("\nUsage: strip [/showonly] directory | EXE name");
    printf("\nStrips EXE files or .reloc sections\n");
    return EXIT_FAILURE;
    }

  ArgCount = 1;
  if (*argv[ArgCount] == '/') {
    ShowOnly = TRUE;
    ArgCount++;
    }

  NamePtr = strchr(argv[ArgCount], '.');
  if (NamePtr != NULL) {
    // It's a filename
    ProcessFile(argv[ArgCount]); 
    }  
  else {    
    // It's a directory 
    DirectoryLoop(argv[ArgCount]);
    printf("\nTotal Space Saved: %d", TotalSpaceSaved);
    }
  return EXIT_SUCCESS;
  }

// Recursively descend all subdirectories calling FileLoop with
// each directory name as the parameter.
void DirectoryLoop(const char *dir) {
  char  NewDir[MAX_PATH];
  WIN32_FIND_DATA SubDir;
  HANDLE SearchHandle;

  sprintf(SubDirName, "%s\\", dir);
  FileLoop(SubDirName);
  strcat(SubDirName, "*.");
  SearchHandle = FindFirstFile(SubDirName, &SubDir);
  if (SearchHandle != INVALID_HANDLE_VALUE) {      
    do {
      if (strcmp(SubDir.cFileName, ".") &&
          strcmp(SubDir.cFileName, "..") &&
          SubDir.dwFileAttributes == FILE_ATTRIBUTE_DIRECTORY) {
        sprintf(NewDir, "%s\\%s", dir, SubDir.cFileName);
        DirectoryLoop(NewDir);
        }
      } while (FindNextFile(SearchHandle, &SubDir));
    FindClose(SearchHandle);
    }
  }

// Find all EXEs in the directory provided.
void FileLoop(const char *dir) {
  char Filter[MAX_PATH], FullPathName[MAX_PATH];
  WIN32_FIND_DATA FileInfo;
  HANDLE SearchHandle;
  
  sprintf(Filter, "%s*.exe", dir);
  SearchHandle = FindFirstFile(Filter, &FileInfo);
  if (SearchHandle != INVALID_HANDLE_VALUE) {
    do {
      sprintf(FullPathName, "%s%s", dir, FileInfo.cFileName);
      ProcessFile(FullPathName);
      } while (FindNextFile(SearchHandle, &FileInfo));
    FindClose(SearchHandle);
    }
  }

// Strips .reloc if all conditions met.
void ProcessFile(const char *FullPathName) {
  IMAGE_DOS_HEADER DosHeader;
  WORD Signature;
  IMAGE_FILE_HEADER PEFileHeader;
  IMAGE_OPTIONAL_HEADER OptionalHeader;
  IMAGE_SECTION_HEADER SectionHeader;
  long FileHeaderPosition;
  WORD Skip, i;
  FILE *fp;
  DWORD SizeOfFile, NewFileLength;
  BYTE *buffer;

  fp = fopen( FullPathName, "r+b");
  if (!fp)
    {
    FAIL("Can't open file.");
    }

  // Read DOS header
  fread(&DosHeader, sizeof(IMAGE_DOS_HEADER), 1, fp);
  if (DosHeader.e_magic != IMAGE_DOS_SIGNATURE) {
    fclose(fp);
    FAIL("Not an executable.");
    }
  if (DosHeader.e_lfarlc < 0x40) {
    fclose(fp);
    FAIL("It's a DOS 16-bit executable.");
    }

  // Read the signature
  fseek(fp, DosHeader.e_lfanew, SEEK_SET);
  fread(&Signature, sizeof(WORD), 1, fp);
  if (Signature != IMAGE_NT_SIGNATURE) {
    fclose(fp);
    FAIL("No NT signature in EXE.");
    }

  // PE File header
  fread(&Skip, sizeof(WORD), 1, fp);
  FileHeaderPosition = ftell(fp);
  fread(&PEFileHeader, IMAGE_SIZEOF_FILE_HEADER, 1, fp);
  if ((PEFileHeader.Characteristics & 
       IMAGE_FILE_RELOCS_STRIPPED) != 0) {
    fclose(fp);
    FAIL("Already stripped.");
  }

  // Read the optional header
  fread(&OptionalHeader, IMAGE_SIZEOF_NT_OPTIONAL_HEADER, 1, fp);

  // Make sure desired address is at least 400000h.
  if (OptionalHeader.ImageBase < 0x400000) {
    fclose(fp);
    FAIL("Incorrect preferred loading address.");
    }

  // Read sections, looking for the .reloc section, which 
  // must be last.
  for (i=0;i<PEFileHeader.NumberOfSections;i++)
    fread(&SectionHeader, IMAGE_SIZEOF_SECTION_HEADER, 1, fp);

  if (strcmp((LPCTSTR)SectionHeader.Name, ".reloc") == 0) {
    fseek(fp, 0, SEEK_END);
    SizeOfFile = ftell(fp);
    if ((SectionHeader.PointerToRawData +
         SectionHeader.SizeOfRawData) != SizeOfFile) {
      fclose(fp); 
      FAIL(".reloc not at end of EXE.");
      }
    printf("\nSaved %d bytes in %s",
           SectionHeader.SizeOfRawData, FullPathName);
    TotalSpaceSaved += SectionHeader.SizeOfRawData;
    if (!ShowOnly) {
      PEFileHeader.NumberOfSections--;
      PEFileHeader.Characteristics |= IMAGE_FILE_RELOCS_STRIPPED;
      buffer = (BYTE *)malloc(SizeOfFile);
      fseek(fp, 0, SEEK_SET);
      fread(buffer, sizeof(BYTE), SizeOfFile, fp);
      memcpy(&buffer[FileHeaderPosition], &PEFileHeader,
             IMAGE_SIZEOF_FILE_HEADER);
      NewFileLength = SizeOfFile - SectionHeader.SizeOfRawData;
      fclose(fp);
      fp = fopen( FullPathName, "wb");
      fwrite(buffer, sizeof(BYTE), NewFileLength, fp);
      free(buffer);
      printf("\nStripped %s", FullPathName);
      }
    }
  else
    {
    FAIL(".reloc was not last section in EXE.");
    }
  fclose(fp);
  }
