// joystick.c

#include "ntddk.h"
#include "gameport.h"
#include "joystick.h"

typedef struct _DEVICE_EXTENSION {
   PDEVICE_OBJECT    TargetDevice;
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

NTSTATUS JoyStickDispatch(IN PDEVICE_OBJECT, IN PIRP);
NTSTATUS JoyStickRead(IN PDEVICE_OBJECT, IN PIRP);
VOID JoyStickUnload(IN PDRIVER_OBJECT);
NTSTATUS JoyStickCompletion(IN PDEVICE_OBJECT, IN PIRP Irp, PVOID);

//-----------------------------------------------------------------
NTSTATUS DriverEntry(IN PDRIVER_OBJECT  DriverObject,
                     IN PUNICODE_STRING RegistryPath)
{
    NTSTATUS           status = STATUS_SUCCESS;
    PDEVICE_EXTENSION  deviceExt;
    PDEVICE_OBJECT     deviceObject, targetDevice;
    PDRIVER_DISPATCH   invalidRequest;
    UNICODE_STRING     deviceName;
    ULONG              i;

    UNREFERENCED_PARAMETER(RegistryPath);

    KdPrint(("JoyStick!DriverEntry: Driver loading\n"));

    //
    // Create an unnamed device object to represent the joystick
    //
                     
    status = IoCreateDevice(DriverObject,sizeof(DEVICE_EXTENSION),
                            NULL, FILE_DEVICE_JOYSTICK, 0, FALSE,
                            &deviceObject);
    if (!NT_SUCCESS(status)) {
        return status;
    }

    deviceObject->Flags |= DO_BUFFERED_IO;

    //
    // Attach the joystick object to the gameport object
    //

    RtlInitUnicodeString(&deviceName, GAMEPORT_DEV_NAME);

    status = IoAttachDevice(deviceObject, &deviceName, &targetDevice);
    if (!NT_SUCCESS(status)) {
        KdPrint(("JoyStick Attach failed (%x)\n", status));
        IoDeleteDevice(deviceObject);
        return status;
    }

    deviceExt = (PDEVICE_EXTENSION)deviceObject->DeviceExtension;
    deviceExt->TargetDevice = targetDevice;  

    //
    // Fill in the dispatch table. Use a generic dispatch routine 
    // for any routines the next driver exposes.
    //

    invalidRequest = DriverObject->MajorFunction[IRP_MJ_READ];

    for (i=0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++) {
        if (targetDevice->DriverObject->MajorFunction[i] 
                    != invalidRequest) {
            DriverObject->MajorFunction[i] = JoyStickDispatch;
        }
    }

    DriverObject->MajorFunction[IRP_MJ_READ] = JoyStickRead;
    DriverObject->DriverUnload               = JoyStickUnload;
    return status;

} // DriverEntry

//-----------------------------------------------------------------
NTSTATUS JoyStickDispatch(IN PDEVICE_OBJECT DevObject, IN PIRP Irp)
{
    PIO_STACK_LOCATION currentIrpStack, nextIrpStack;
    PDEVICE_EXTENSION deviceExt = DevObject->DeviceExtension;

    currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
    nextIrpStack = IoGetNextIrpStackLocation(Irp);
    
    KdPrint(("JoyStickDispatch %x\n", 
             currentIrpStack->MajorFunction));

    //
    // Block any write requests to the GamePort device
    //

    if (currentIrpStack->MajorFunction == IRP_MJ_WRITE) {
        KdPrint(("JoyStickDispatch: Blocking a write request\n")); 
        Irp->IoStatus.Status = STATUS_ACCESS_DENIED;
        Irp->IoStatus.Information = 0;
        IoCompleteRequest(Irp, IO_NO_INCREMENT);
        return STATUS_ACCESS_DENIED;
    }    

    //
    // Pass all other requests on to next lower driver without
    // modification -- copy stack parameters to next stack, set
    // I/O completion routine and call next driver.
    //

    RtlMoveMemory(nextIrpStack, currentIrpStack,
                  sizeof(IO_STACK_LOCATION));
  
    IoSetCompletionRoutine(Irp, JoyStickCompletion, NULL, TRUE,
                           TRUE, TRUE); 

    return IoCallDriver(deviceExt->TargetDevice, Irp);

} // JoyStickDispatch

//-----------------------------------------------------------------
NTSTATUS JoyStickRead(IN PDEVICE_OBJECT DevObject, IN PIRP Irp)
{
    PIO_STACK_LOCATION currentIrpStack, nextIrpStack;
    PDEVICE_EXTENSION deviceExt = DevObject->DeviceExtension;

    KdPrint(("JoyStickRead\n"));

    currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
    nextIrpStack = IoGetNextIrpStackLocation(Irp);

    //
    // Validate and reset read parameters (gameport expects one
    // byte read, but joystick returns JOY_DATA).
    //
    
    if (currentIrpStack->Parameters.Read.Length
                < sizeof(JOY_DATA)) {
        Irp->IoStatus.Status = STATUS_BUFFER_TOO_SMALL;
        Irp->IoStatus.Information = 0;
        return STATUS_BUFFER_TOO_SMALL;
    }

    //
    // Pass to next driver -- set up an I/O completion routine so
    // we can take another crack at processing read requests.
    //

    RtlMoveMemory(nextIrpStack, currentIrpStack,
                  sizeof(IO_STACK_LOCATION));
    nextIrpStack->Parameters.Read.Length = 1;
    nextIrpStack->Parameters.Read.Key = 0;     // private use 

    IoSetCompletionRoutine(Irp, JoyStickCompletion, NULL, TRUE,
                           TRUE, TRUE); 

    return IoCallDriver(deviceExt->TargetDevice, Irp);

} // JoyStickRead

//-----------------------------------------------------------------
NTSTATUS JoyStickCompletion(IN PDEVICE_OBJECT DevObject, 
                            IN PIRP Irp, IN PVOID Context)
{
    PIO_STACK_LOCATION irpStack = IoGetCurrentIrpStackLocation(Irp);
    PDEVICE_EXTENSION deviceExt = DevObject->DeviceExtension;
    JOY_DATA joyData;
    UCHAR data;

    KdPrint(("JoyStickRead\n"));

    if (irpStack->MajorFunction == IRP_MJ_READ) {
        if (NT_SUCCESS(Irp->IoStatus.Status)) {
            data = *(PUCHAR)(Irp->AssociatedIrp.SystemBuffer);
            joyData.Device1Button1 = (data & 0x10) ? 0 : 1;
            joyData.Device1Button2 = (data & 0x20) ? 0 : 1;
            joyData.Device2Button1 = (data & 0x40) ? 0 : 1;
            joyData.Device2Button2 = (data & 0x80) ? 0 : 1;

            RtlMoveMemory(Irp->AssociatedIrp.SystemBuffer, 
                          &joyData, sizeof(JOY_DATA));
            Irp->IoStatus.Information = sizeof(JOY_DATA);
        }    
    }

    return STATUS_SUCCESS;

} // JoyStickCompletion

//-----------------------------------------------------------------
VOID JoyStickUnload(IN PDRIVER_OBJECT DriverObject)
{
    PDEVICE_OBJECT   deviceObject = DriverObject->DeviceObject;
    PDEVICE_EXTENSION deviceExt = deviceObject->DeviceExtension;

    KdPrint(("JoyStickUnload\n"));

    //
    // First detach from lower device, then delete this device.
    //
    
    IoDetachDevice(deviceExt->TargetDevice);
    IoDeleteDevice(deviceObject);

} // JoyStickUnload



